local game = {}

require("utils.b2d.access")
require("utils.b2d.draw")
love.physics.setMeter(20)

game.oo = require("utils.pure.oo")
game.progress = require("core.progress")

game.cat =
{
  none = 0,
  player = 1,
  enemy = 2,
  bullet = 4,
  wall = 8,
  platform = 16,
  env = 24,
  ladder = 32,
  debris = 64,
  pdefault = 0xffff - 1,
  default = 0xffff,
}
game.groups =
{
  none = 0,
  player = -1,
  enemy = -2,
  bullet = -3,
  landmine = -4,
  debirs = -5,
}
game.actions =
{
  'up', 'down', 'left', 'right', 'jump', 'shoot', 'start',
}
game.glyphs =
{
  ["left"] = "←", ["right"] = "→", ["up"] = "↑", ["down"] = "↓",
  ["dpleft"] = "⇦", ["dpright"] = "⇨", ["dpup"] = "⇧", ["dpdown"] = "⇩",
  ["return"] = "↵", ["kpenter"] = "KP↵",
}

game.newLevel = require("core.level")

game.clampx = 8*2
game.clampy = 8*2.5
game.minlives = 5

function game.newSession()
  return
  {
    level = "1",
    seed = math.random(1, 1000),
    replay = false,
    online = false,
    invite = false,
    lobby = nil,
    lives = game.minlives,
    joined = 0,
    p1c = nil,
    p2c = nil,
  }
end

function game.load(session)
  math.randomseed(session.seed)
  local w, h = love.window.getMode()
  game.lvl = game.newLevel(session)
  game.lvl.init()
  return game.lvl
end

function game.unload(lvl)
  -- very quick and dirty hack for love.js
  if love.audio and app.ops.os ~= "Web" then
    love.audio.stop()
  end
  while #lvl.session > 0 do
    table.remove(lvl.session)
  end
  lvl.destroy()
  if lvl == game.lvl then
    game.lvl = nil
    game.socket = nil
  end
end

function game.flush(lvl)
  local d = lvl.data
  local n = lvl.title
  local p = game.progress.data
  p.elapsed = math.floor(p.elapsed + d.elapsed)
  -- best score
  local top = math.max(p["score"..n] or 0, d.score)
  p["score"..n] = top
  if top == d.maximum then
    p["arena"..n] = true
  end
  if d.complete then
    -- extra life?
    if top == d.maximum then
      local s = lvl.session
      s.lives = math.max(s.lives, game.minlives) + 1
    end
    -- best time
    local best = p["time"..n] or 0
    if best == 0 or best > d.elapsed then
      p["time"..n] = d.elapsed
    end
  end
end

-- nothing too heavy
function game.update(lvl, dt)
  local d = lvl.data
  if not d.over then
    if not game.poll(lvl, d.tick) then
      return
    end
  end
  -- update the simulation
  d.tick = d.tick + 1
  d.elapsed = d.elapsed + dt
  lvl.world:update(dt, 10, 10)
  lvl.broadcast("update", dt)
  lvl.destroyQueue()
  game.updateCamera(lvl, dt)
end

function game.updateCamera(lvl, dt)
  --local view = lvl.view
  --local vx, vy, vw, vh = view:getScene()
  --local tx, ty = vx, -vy
  --view:setScene(tx, -ty, vw, vh)
  local d = lvl.data
  if d.duration then
    d.duration = d.duration - dt
    if d.duration <= 0 then
      d.duration = nil
    end
  end

  local w, h = lvl.getDimensions()
  if w > 35 or h > 20 then
    lvl.updateRange(lvl.data.p1)
    lvl.updateRange(lvl.data.p2)
    --for obj in pairs(lvl.dynamic) do
    for _, obj in ipairs(lvl.dynamic) do
      obj:updateRefs()
    end
  end
end

function game.screenshake(lvl, duration)
  lvl.data.duration = duration
end

function game.clampCamera(lvl, rw, rh, px, py, cx, cy)
  -- player
  cx, cy = cx or px, cy or py
  local mw, mh = lvl.getDimensions()
  mw, mh = mw*8, mh*8
  local h, v = game.clampx, game.clampy
  local dx, dy = cx - px, cy - py
  if dx > h then
    cx = cx - (dx - h)
  elseif dx < -h then
    cx = cx + (-dx - h)
  end
  if dy > v then
    cy = cy - (dy - v)
  elseif dy < -v then
    cy = cy + (-dy - v)
  end
  -- bounds
  if mw > rw then
    cx = math.max(cx, rw/2)
    cx = math.min(cx, mw - rw/2)
  else
    cx = mw/2
  end
  if mh > rh then
    cy = math.max(cy, rh/2)
    cy = math.min(cy, mh - rh/2)
  else
    cy = mh/2
  end
  return cx, cy
end

function game.drawView(lvl, vx, vy, vw, vh, cx, cy)
  -- shake
  local d = lvl.data
  if d.duration then
    cx = cx + math.crandom(-1, 1)
    cy = cy + math.crandom(-1, 1)
  end
  -- round
  cx = math.floor(cx + 0.5)
  cy = math.floor(cy + 0.5)
  vw = math.floor(vw)
  vh = math.floor(vh)
  vw = vw - vw%2
  vh = vh - vh%2
  local view = lvl.view
  view:setScene(cx, -cy)
  view:setBounds(vx, vy, vw, vh)
  view:draw()
end

function game.draw(lvl)
  lvl.draw()
  
  -- debugdraw
  if app.arg.debugdraw or app.ops.debugdraw then
    local view = lvl.view
    local vw, vh = view:getDimensions()
    local mw, mh = lvl.getDimensions()
    mw = mw*8*app.s
    mh = mh*8*app.s
    local lg = love.graphics
    lg.reset()
    lg.setColor(1, 0, 0, 1)
    if mw > 35 or mh > 20 then
      local cx, cy, cw, ch = view:getScene()
      lg.translate(-cx*app.s, cy*app.s)
      lg.translate(vw/2, vh/2)
    else
      lg.translate(vw/2 - mw/2, vh/2 - mh/2)
    end
    lg.scale(app.s, app.s)
    lg.setLineWidth(0.25)
    lvl.world:draw()
  end
end

function game.newSpritesheet(img, cols, rows)
  local quad = {}
  local w, h = img:getDimensions()
  local tw, th = w/cols, h/rows
  for r = 1, rows do
    for c = 1, cols do
      local i = (r - 1)*cols + c
      local x, y = (c - 1)*tw, (r - 1)*th
      quad[i] = love.graphics.newQuad(x, y, tw, th, w, h)
    end
  end
  return quad
end

function game.newAnimation(a, n)
  local step = 1/app.ops.framerate --*2
  a.steps = a.steps or 2
  a.rate = a.rate or step*a.steps
  a.total = a.frames*a.rate
  a.anim = a
  a.sheet = game.image[n]
  a.quad = game.quad[n]
  assert(a.quad, n)
  local _, _, w, h = a.quad[1]:getViewport()
  a.ox = a.ox or -w/2
  a.oy = a.oy or -h/2
  return a
end

function game.newAnimations(defs)
  local t = {}
  for k, v in pairs(defs) do
    if type(v) == "table" then
      v.ox = v.ox or defs.ox
      v.oy = v.oy or defs.oy
      t[k] = game.newAnimation(v, v.file or defs.file)
    end
  end
  return t
end

function game.event(lvl, ...)
  local s = lvl.session
  table.insert(s, { lvl.data.tick, ... })
  assert(not s.replay)
end

function game.input(lvl, p, state2)
  local changed = true
  local d = lvl.data
  local state1 = (p == "p1") and d.p1a or d.p2a
  -- save input for replays and stuff
  local s = lvl.session
  for _, a in ipairs(game.actions) do
    local on1 = state1[a] == true
    local on2 = state2[a] == true
    if on1 ~= on2 then
      game.event(lvl, p, a, on2)
      changed = true
    end
    state1[a] = on2
  end
  return changed
end

function game.send(tick, state)
  local lo = tick%256
  local hi = (tick - lo)/256
  local msg = string.char(hi, lo)
  if state then
    local bits = 0
    for i, a in ipairs(game.actions) do
      local n = state[a] and 1 or 0
      bits = bits + 2^(i - 1)*n
    end
    msg = msg..string.char(bits)
  end
  game.socket:send(msg)
end

function game.receive(msg)
  local hi, lo = msg:byte(1, 2)
  local tick = hi*256 + lo
  local state = nil
  if msg:len() >= 3 then
    state = {}
    local bits = msg:byte(3)
    for i, a in ipairs(game.actions) do
      local rest = bits%2
      state[a] = (rest == 1)
      bits = (bits - rest)/2
    end
  end
  return tick, state
end

local steps = 8
local resync = 1/(steps/2)

function game.poll(lvl, tick)
  local d = lvl.data
  local s = lvl.session

  -- store input from controllers
  if s.replay then
    -- replay
    local last = s[#s]
    if not last or tick > last[1] then
      d.over = 0
    end
  elseif game.socket then
    -- online
    local own, peer = "p1", "p2"
    if s.p1c == "socket" then
      own, peer = peer, own
    end
    d.cached = d.cached or {}
    for id, device in pairs(app.attached) do
      for b, a in pairs(app.bindings[id]) do
        if not d.cached[a] and app.input.isDown(device, b) then
          d.cached[a] = true
          s[own.."c"] = id
        end
      end
    end
    if not d.sent[tick] then
      -- send input to socket
      local changed = game.input(lvl, own, d.cached)
      if tick == 1 then
        changed = true
      end
      for i = 0, steps - 1 do
        d.sent[tick + i] = d.cached
      end
      game.send(tick, changed and d.cached or nil)
      --game.send(tick, pa)
      d.cached = nil
    end
    -- grab input from socket
    repeat
      local msg = game.socket:receive()
      if msg then
        local t, state = game.receive(msg)
        --if t <= tick + 50 then
          state = state or true
          for i = 0, steps - 1 do
            d.received[t + i] = state
          end
        --end
      end
    until not msg
    -- not enough input received
    local now = love.timer.getTime()
    if not d.received[tick] then
      if now - d.sync > resync then
        d.sync = now
        d.resend = d.resend + 1
        local t = tick - tick%steps - steps
        if t >= 0 then
          game.send(t, d.sent[t])
        end
        if d.resend >= 50 then
          -- no response from peer
          d.over = 3
          d.message = app.lang.connectionfailed
        elseif d.resend >= 5 then
          -- connection stalled
          d.message = app.lang.wait
        end
      end
      return false
    end
    d.sync = now
    d.resend = 0
    if type(d.received[tick]) == "table" then
      d.message = nil
      game.input(lvl, peer, d.received[tick])
    end
  else
    -- local
    local p1a, p2a = {}, {}
    if s.joined == 1 then
      -- all controllers are assigned to the same player
      for id, device in pairs(app.attached) do
        for b, a in pairs(app.bindings[id]) do
          if not p1a[a] and app.input.isDown(device, b) then
            p1a[a] = true
            s.p1c = id
          end
        end
      end
    else
      -- each controller is assigned to a different player
      for id, device in pairs(app.attached) do
        for b, a in pairs(app.bindings[id]) do
          if app.input.isDown(device, b) then
            if s.p1c == nil and s.p2c ~= id then
              s.p1c = id
            elseif s.p2c == nil and s.p1c ~= id then
              s.p2c = id
            end
            if id == s.p1c then
              p1a[a] = true
            elseif id == s.p2c then
              p2a[a] = true
            end
          end
        end
      end
    end
    -- events
    game.input(lvl, "p1", p1a)
    game.input(lvl, "p2", p2a)
  end

  -- read from session
  for i, e in ipairs(s) do
    if e[1] == tick then
      local state = d.p1a
      if e[2] == "p2" then
        state = d.p2a
      end
      state[ e[3] ] = e[4]
    elseif e[1] > tick then
      break
    end
  end

  return true
end

return game