local lg = love.graphics
return function()
  print("loading started")
  
  if app.arg.devmode then
    print("entering developer mode")
    -- strict
    require("utils.pure.strict")
    strict_declare_globals("game","ltn12")
    -- pipeline
    for fn in pairs(app.arg) do
      if app.exists("/tools/"..fn..".lua") then
        print("command:"..fn)
        require("tools."..fn)
      end
    end
  end
  
  print("preparing environment")
  local ops = app.ops
  local src = love.filesystem.getSource()
  if src:match("^.+%.(.+)$") == "love" then
    if app.exists("utils/maker/") then
      local maker = require("utils.maker.main")
      if maker then
        ops.build = maker.getComment(src) or ops.build
      end
    end
  end
  local v1, v2, v3 = love.getVersion()
  ops.love = string.format("%d.%d.%d", v1, v2, v3)
  ops.os = love.system.getOS()
  if ops.os == "Android" or ops.os == "iOS" then
    app.arg.mobile = true
  end

  app.mkdir(ops.version)
  
  love.resize(app.w, app.h)
  
  print('loading core')
  -- table
  require("utils.pure.table")
  require("utils.pure.random")
  -- scene graph
  love.scene = require("utils.scene.main")
  -- core
  game = require("core.game")
  require("core.graphics")

  print("loading locales")
  app.text = require("utils.pure.text")
  --- load locales
  local locales = {}
  local files = app.getdir("/lang/")
  for _, fn in ipairs(files) do
    local lc = fn:match(".-([^\\/]-)%.?[^%.]*$")
    local t = require("lang."..lc)
    if type(t) == "table" then
      t.align = "left"
      if lc == "ar" or lc == "he" then
        for k, v in pairs(t) do
          t[k] = app.text.rtl(v)
        end
        t.align = "right"
      end
      locales[lc] = t
    end
    coroutine.yield(0, 1)
  end
  --- fallback
  for lc, t in pairs(locales) do
    local base = lc == "en" and app.text or locales[t._base or "en"]
    setmetatable(t, { __index = base })
  end
  if locales.ar then
    locales.ar.number = app.text.numberha
  end
  -- select preferred language
  print("selected locale:"..ops.language)
  app.lang = locales[ops.language] or locales.en
  app.text.locales = locales

  print("initializing input")
  app.input = {}
  app.attached = {}
  app.polled = {}
  app.bindings = {}
  if not app.arg.reset then
    local bind = app.load(ops.version.."/bindings.lua")
    app.bindings = bind or app.bindings
  end
  function app.input.getid(device)
    for k, v in pairs(app.attached) do
      if device == v then
        return k
      end
    end 
  end
  function app.input.add(device, name)
    -- handle duplicate joystick ids
    local i = 1
    local id = name
    while app.attached[id] do
      i = i + 1
      id = id..'('..i..')'
    end
    print("attached:"..id)
    app.attached[id] = device
    app.polled[id] = {}
    if not app.bindings[id] then
      if device == love.keyboard then
        local p = require("defs.keymap")
        for k1, v in pairs(p) do
          local k2 = device.getKeyFromScancode(k1)
          p[k1] = nil
          p[k2] = v
        end
        app.bindings[id] = p
      else
        app.bindings[id] = require("defs.joymap")
      end
    end
  end
  function app.input.remove(device)
    local id = app.input.getid(device)
    print("detached:"..id)
    app.attached[id] = nil
    app.polled[id] = nil
  end
  function app.input.isDown(d, b)
    return d == love.keyboard and d.isDown(b) or d:isDown(b)
  end
  function app.input.poll()
    for id, device in pairs(app.attached) do
      local bind = app.bindings[id]
      for b, a in pairs(bind) do
        local p = app.polled[id]
        local on = app.input.isDown(device, b)
        if on ~= p[b] then
          p[b] = on
          if on then
            app.callstate(a, id)
          end
        end
      end
    end
  end

  if love.keyboard then
    app.input.add(love.keyboard, "Keyboard")
    function love.textinput(text)
      app.callstate("textinput", text)
    end
    function love.keypressed(key, c, r)
      app.callstate("keypress", key, c, r)
      if key == "f5" then
        if love.keyboard.isDown("lalt", "ralt") then
          local dir = love.filesystem.getSaveDirectory()
          love.system.openURL("file://"..dir.."/"..app.ops.version)
          love.event.quit()
          return
        end
      end
      if key == "return" or key == "kpenter" then
        -- fulscreen toggle?
        if love.keyboard.isDown("lalt", "ralt") then
          local on = not love.window.getFullscreen()
          if not love.window.setFullscreen(on) then
            return
          end
          love.mouse.setVisible(not on)
          app.ops.fullscreen = on
          --app.callstate("reset")
          local dw, dh = love.window.getDesktopDimensions(ops.display)
          local w, h = lg.getDimensions()
          if not on and dh - h < 25 then
            love.window.maximize()
          end
          love.resize(w, h)
        end
      end
    end
  end

  function love.joystickadded(device)
    local name = device:getName()
    app.input.add(device, name)
    local id = app.input.getid(device)
    app.callstate("attach", id)
  end
  function love.joystickremoved(device)
    local id = app.input.getid(device)
    app.input.remove(device)
    app.callstate("detach", id)
  end

  if love.joystick then
    local joysticks = love.joystick.getJoysticks()
    for _, device in ipairs(joysticks) do
      local name = device:getName()
      app.input.add(device, name)
    end
    
    require("utils.love.joystick")
    local reg = debug.getregistry()
    if reg.Joystick then
      local isDown = reg.Joystick.isDown
      function reg.Joystick.isDown(device, b)
        local id = app.input.getid(device)
        local n = app.ops.nullzone[id] or 0.5
        n = n*0.9 + 0.1
        return reg.Joystick.isDownEx(device, b, n)
      end
    end
  end

  print("loading fonts")
  local fonts = {}
  app.fonts = fonts
  fonts.mini = love.graphics.newFont("/fonts/arena.ttf", 9)
  if app.exists("/fonts/ydwebdotum.ttf") and app.exists("/fonts/zuixiangsu.ttf") then
    fonts.hangul = love.graphics.newFont("/fonts/ydwebdotum.ttf", 9)
    fonts.chinese = love.graphics.newFont("/fonts/zuixiangsu.ttf", 10)
    fonts.mini:setFallbacks(fonts.hangul, fonts.chinese)
  end
  fonts.minib = love.graphics.newImageFont("/fonts/minib.png", "0123456789.,:;/\\💀❤🕒")
  fonts.intv = love.graphics.newImageFont("/fonts/intv.png", "0123456789")
  for k, v in pairs(fonts) do
    v:setFilter(ops.filter, ops.filter)
  end
  --[[
  if app.arg.devmode then
    for lc, text in pairs(app.text.locales) do
      for k, v in pairs(text) do
        v = v:gsub("\n", "")
        if not app.fonts.mini:hasGlyphs(v) then
          error("missing glyph:"..lc..' ('..k..')')
        end
      end
    end
  end
  ]]
  
  print("loading sounds")
  app.sfx = {}
  app.voice = {}
  if love.audio then
    local sfx = app.getdir("/sfx/", "flac")
    for _, fn in ipairs(sfx) do
      fn = "/sfx/"..fn
      print("loading:"..fn)
      local n = fn:match(".-([^\\/]-)%.?[^%.]*$")
      app.sfx[n] = love.sound.newSoundData(fn)
      coroutine.yield()
    end
    local voice = app.getdir("/voice/", "wav")
    for _, fn in ipairs(voice) do
      fn = "/voice/"..fn
      print("loading:"..fn)
      local n = fn:match(".-([^\\/]-)%.?[^%.]*$")
      app.voice[n] = love.sound.newSoundData(fn)
      coroutine.yield()
    end
  end
  
  function app.say(s, p)
    p = p or 10
    if love.audio then
      if app.speak and app.speak:isPlaying() then
        if p >= app.priority then
          app.speak:stop()
          app.speak = nil
        else
          return
        end
      end
      local src = s
      if type(s) == "string" then
        src = love.audio.newSource(app.voice[s], "static")
        src:play()
      end
      src:setVolume(ops.vvolume)
      app.speak = src
      app.priority = p
      return src
    end
  end

  function app.playSound(s, v, i, p)
    if love.audio then
      v = v or 1
      i = i or 1
      p = p or 0
      local src = s
      if type(s) == "string" then
        src = love.audio.newSource(app.sfx[s], "static")
        src:play()
      end
      src:setVolume(ops.svolume*v)
      src:setPitch(i)
      --if ops.channels > 1 and src:getChannelCount() == 1 then
        src:setPosition(0, p, 0)
      --end
      return src
    end
  end

  function app.playMusic(n, v)
    if not love.audio then
      return
    end
    v = v or 1
    local src = app.music
    if n ~= app.track then
      app.stopMusic()
      local fn = "music/"..n..".flac"
      src = love.audio.newSource(fn, "stream")
      src:setLooping(true)
      src:play()
      app.music = src
      app.track = n
    end
    if src then
      src:setVolume(ops.mvolume/4*v)
    end
  end
  function app.stopMusic()
    if app.music then
      app.music:stop()
      app.music = nil
      app.track = nil
    end
  end

  print("loading images")
  game.image = {}
  game.quad = {}
  local gfx = app.getdir("/gfx/tile/", "png")
  for _, fn in ipairs(gfx) do
    fn = "/gfx/tile/"..fn
    print("loading:"..fn)
    local n1 = fn:match(".-([^\\/]-)%.?[^%.]*$")
    local n2, c, r = n1:match("(.-)(%d+)x(%d+)$")
    local img = love.graphics.newImage(fn)
    img:setFilter(ops.filter, ops.filter)
    local n = n2 or n1
    game.image[n] = img
    if c and r then
      game.quad[n] = game.newSpritesheet(img, c, r)
    end
    coroutine.yield()
  end
  --game.image.avatar:setFilter("linear", "linear")

  print("loading objects")
  game.obj = {}
  local files = app.getdir("/obj/", "lua")
  for i, v in ipairs(files) do
    local n = v:gsub("%.lua", "")
    local c = require("obj."..n)
    c.type = n
    local a = rawget(c, "animations")
    if a then
      local q = game.newAnimations(a)
      rawset(c, "animations", q)
    end
    game.obj[n] = c
  end
  
  print("loading shaders")
  app.shaders = {}
  local files = app.getdir("/shaders/", "frag")
  for i, fn in ipairs(files) do
    local n = string.match(fn, ".-([^\\/]-)%.?[^%.]*$")
    app.shaders[n] = love.graphics.newShader("shaders/"..fn)
  end
  app.shaders.default = app.shaders[app.ops.shader]

  if not app.arg.devmode and not app.arg.mobile then
    if app.exists("/crash.txt") then
      love.window.minimize()
      local buttons = { "Cancel", "Log", "OK", enterbutton = 3, escapebutton = 1 }
      local title = "Crash detected"
      local message = "Would you like to play in safe mode with limited features?"
      local chosen = love.window.showMessageBox(title, message, buttons, "info", false)
      if chosen == 1 then
        love.event.quit()
      elseif chosen == 2 then
        local dir = love.filesystem.getSaveDirectory()
        love.system.openURL("file://"..dir.."/crash.txt")
      else
        love.window.restore()
        app.ops.audio = false
      end
    end
    if app.arg.error then
      error("error reporting test")
    end
    app.report(app.ops.build)
  end

  return -1
end