local cycles =
{
  none = function(i,n) return math.min(i, n) end,
  reverse = function(i,n) return math.max(n - i, 0) end,
  loop = function(i,n) return i%(n + 1) end,
  rloop = function(i,n) return n - i%(n + 1) end,
  bounce = function(i,n) return n - math.abs(i%(n*2) - n) end,
  rbounce = function(i,n) return math.abs(i%(n*2) - n) end,
}

local class = game.oo.class()

function class:init(lvl, x, y, z)
  self.sprite = lvl.layers[z]:newSprite(x, y)
  self.lvl = lvl
  self.refs = 1
  self.ix = x
  self.iy = y
  self.mode = true
end

function class:destroy()
  self:stopLoop()
  self.sprite:destroy()
  self.sprite = nil
  self.lvl = nil
  self.refs = nil
  self.ix = nil
  self.iy = nil
  self.mode = nil
end

function class:getType()
  return self.type
end

function class:isStatic()
  return true
end

function class:ref()
  self.refs = self.refs + 1
  if self.refs == 1 then
    self:setAwake(true)
  end
end

function class:unref()
  if self.refs > 0 then
    self.refs = self.refs - 1
    if self.refs == 0 then
      self:setAwake(false)
    end
  end
end

function class:updateRefs()
  if self:inRange() then
    if self.refs == 0 then
      self:ref()
    end
  else
    self:unref()
  end
end

function class:reposition()
  self:setPosition(self.ix, self.iy)
  return self.ix, self.iy
end

function class:inRange(x, y)
  local cx, cy = self:getTile(x, y)
  for _, r in pairs(self.lvl.range) do
    if cx >= r[1] and cx <= r[3] and cy >= r[2] and cy <= r[4] then
      return true
    end
  end
  return false
end

function class:getTile(x, y)
  if not x or not y then
    x, y = self:getPosition()
  end
	return math.floor(x/8 + 0.5), math.floor(y/8 + 0.5)
end

function class:playSound(n, v)
  local w = self:calculateVolume()
  v = (v or 1)*(w or 0)
  if v > 0 then
    return app.playSound(n, v)
  end
end

function class:loopSound(n, v)
  local src = app.playSound(n, 0)
  if src then
    self:updateSound(src, v)
    src:setLooping(true)
    self.loop = src
  end
  return src
end

function class:stopLoop()
  if self.loop then
    self.loop:stop()
    self.loop = nil
  end
end

function class:calculateVolume()
  local x1, y1 = self:getPosition()
  local v
  local d = self.lvl.data
  local c = { d.cx, d.cy, d.cx2, d.cy2 }
  for i = 1, #c, 2 do
    local x2, y2 = c[i], c[i + 1]
    local dx = math.abs(x2 - x1)/(8*20)
    local dy = math.abs(y2 - y1)/(8*10)
    dx = 1 - math.min(dx, 1)
    dy = 1 - math.min(dy, 1)
    local v2 = dx*dy
    if not v or v2 > v then
      v = v2
    end
  end
  return v
end

function class:updateSound(src, v)
  if not src then
    return
  end
  local w = self:calculateVolume()
  v = (v or 1)*(w or 0)
  src:setVolume(app.ops.svolume*v)
end

function class:pause()
  if self.loop then
    self.loop:pause()
  end
end

function class:resume()
  if self.loop then
    self.loop:play()
  end
end

function class:call(fn, ...)
  local func = self[fn]
  if type(func) == "function" then
    return func(self, ...)
  end
end

function class:setGraphic(n, i, ox, oy)
  if not n then
    self.sprite:setGraphic()
  else
    ox = ox or 0
    oy = oy or 0
    local img = game.image[n]
    if i then
      local quad = game.quad[n][i]
      local _, _, vw, vh = quad:getViewport()
      self.sprite:setGraphic(img, quad, -vw/2 + ox, -vh/2 + oy)
    else
      local iw, ih = img:getDimensions()
      self.sprite:setGraphic(img, -iw/2 + ox, -ih/2 + oy)
    end
  end
end

function class:getPosition()
  return self.sprite:getPosition()
end

function class:setPosition(x, y)
  return self.sprite:setPosition(x, y)
end

function class:getScale()
  return self.sprite:getScale()
end

function class:setScale(sx, sy)
  self.sprite:setScale(sx, sy)
end

function class:setColor(...)
  self.sprite:setColorRGB(...)
end

function class:getColor()
  return self.sprite:getColorRGB()
end

function class:setAwake(on)
  if self.nosleep then
    return
  end
  on = on and self.mode or false
  self.sprite:setVisible(on)
end

function class:toggle(on)
  self.mode = on
  self.sprite:setVisible(on)
end

function class:spawn(n, dx, dy, ...)
  local x, y = self:getPosition()
  return self.lvl.newObject(n, x + dx, y + dy, ...)
end

function class:debree(n, dx, dy, r, g, b)
  for i = 1, n do
    local obj = self:spawn("debris", dx, dy)
    obj:setColor(r, g, b)
  end
end

function class:setAnimation(a)
  self.animation = a
end

function class:sync()
  local a = self.animation
  if a then
    local e = self.elapsed or self.lvl.data.elapsed
    local i = math.floor(e/a.rate)
    local f = cycles[a.cycle](i, a.frames - 1)
    f = f + a.start
    local b = a.anim
    self.sprite:setGraphic(b.sheet, b.quad[f], b.ox, b.oy)
  end
end

return class