local cat = game.cat

local states = {}

states.standing = {}
function states.standing:enter(obj)
  obj.canclimb = true
  obj.canshoot = true
  obj:setVelocity(0, 0)
end
function states.standing:update(obj, dt)
  local xm, ym, jump, shoot = obj:getMovement()
  if xm ~= 0 then
    obj:setState("running")
  elseif obj.elapsed > 8 and not shoot then
    obj:setState("waving")
  end
end
function states.standing:fall(obj)
  obj:setState("falling")
end
function states.standing:climb(obj)
  obj:setState("climbing")
end
function states.standing:leap(obj)
  local xm, ym = obj:getMovement()
  if ym > 0 then
    obj:setState("backflip")
  elseif ym < 0 then
    if obj.candropdown then
      obj:setState("dropdown")
    end
  else
    obj:setState("jumping")
  end
end
function states.standing:shoot(obj)
  local sx, sy = obj:getScale()
  obj:shoot(sx*2, -1, sx, 0)
end

states.hanging = {}
function states.hanging:enter(obj)
  obj.body:setGravityScale(0)
  obj.body:setFilter(cat.player, cat.pdefault - cat.platform, 0)
  obj:setVelocity(0, 0)
end
function states.hanging:exit(obj)
  obj.body:setGravityScale(1)
  obj.body:setFilter(cat.player, cat.pdefault, 0)
  obj:setVelocity(0, 1)
end
function states.hanging:update(obj, dt)
  obj.elapsed = obj.elapsed - dt
  obj.grounded = false
  local x, y = obj:getPosition()
  x = x - 4
  x = math.floor(x/8 + 0.5)*8
  x = x + 4
  obj:setPosition(x, y)
  local xm, ym = obj:getMovement()
  if xm ~= 0 or not obj:isNearLadder() then
    obj:setState("falling")
  elseif ym ~= 0 then
    obj:setState("climbing")
  end
end
function states.hanging:climb(obj)
  obj:setState("climbing")
end
function states.hanging:leap(obj)
  obj:setState("falling")
end
function states.hanging:hitground(obj, yv)
  obj:setState("standing")
end

states.climbing = {}
function states.climbing:enter(obj)
  obj:loopSound("climb")
  obj.body:setGravityScale(0)
  obj.body:setFilter(cat.player, cat.pdefault - cat.platform, 0)
end
function states.climbing:exit(obj)
  obj.body:setGravityScale(1)
  obj.body:setFilter(cat.player, cat.pdefault, 0)
  obj:stopLoop()
end
function states.climbing:update(obj, dt)
  obj:updateSound(obj.loop)
  if obj.loop then
    local p = obj.loop:getPitch()
    p = math.min(p + 0.5*dt, 2)
    obj.loop:setPitch(p)
  end
  local x, y = obj:getPosition()
  x = x - 4
  x = math.floor(x/8 + 0.5)*8
  x = x + 4
  obj:setPosition(x, y)
  local xm, ym = obj:getMovement()
  obj:setVelocity(0, -ym*50)
  if not obj:isNearLadder() then
    if ym > 0 then
      obj:setVelocity(0, 0)
      obj:setState("standing")
    else
      obj:setState("falling")
    end
  elseif xm ~= 0 then
    obj:setVelocity(0, 0)
    obj:setState("falling")
  elseif ym == 0 then
    local e = obj.elapsed
    obj:setState("hanging")
    obj.elapsed = e
  end
  if obj.elapsed >= 0.1 then
    obj.canclimb = false
  end
end
function states.climbing:fall(obj)
  obj:setState("falling")
end
function states.climbing:leap(obj)
  obj:setState("falling")
end
function states.climbing:hitground(obj, yv)
  obj:setState("standing")
end

states.waving = {}
function states.waving:enter(obj)
  local n = table.random({ "hey", "whereareyou", "overhere", "ifeelasleep" })
  app.say(n, 3)
end
function states.waving:update(obj, dt)
  local xm, ym = obj:getMovement()
  if xm ~= 0 then
    obj:setState("running")
  elseif obj.elapsed > obj.animation.total*3 then
    obj:setState("standing")
  end
end
function states.waving:fall(obj)
  obj:setState("falling")
end

states.running = {}
function states.running:enter(obj)
  local fn = (obj.name == "p1") and "step" or "step2"
  obj:loopSound(fn)
end
function states.running:exit(obj)
  obj:stopLoop()
end
function states.running:update(obj, dt)
  obj:updateSound(obj.loop)
  obj.canshoot = true
  local xv, yv = obj:getVelocity()
  local xm, ym = obj:getMovement()
  local sx, sy = obj:getScale()
  obj:setVelocity(50*xm, yv)
  if xm == -1 or xm == 1 then
    obj:setScale(xm, 1)
  end
  if xm == 0 then
    obj:setState("standing")
  end
end
function states.running:fall(obj)
  obj:setState("falling")
end
function states.running:leap(obj)
  local xm, ym, jump = obj:getMovement()
  if ym < 0 and jump then
    obj:setState("rolling")
  else
    obj:setState("jumping")
  end
  return true
end
function states.running:shoot(obj)
  local sx, sy = obj:getScale()
  obj:shoot(sx*2, -1, sx, 0)
end

states.backflip = {}
function states.backflip:enter(obj)
  obj.canshoot = false
  local xv, yv = obj:getVelocity()
  obj:setVelocity(xv, -240)
  obj:playSound("backflip")
end
function states.backflip:update(obj, dt)
  local xv, yv = obj:getVelocity()
  local sx, sy = obj:getScale()
  obj:setVelocity(-50*sx, yv)
  if yv > 0 then
    obj:setState("backfall")
  end
end
function states.backflip:hitground(obj, yv)
  obj:setState("landing")
end

states.backfall = {}
function states.backfall:enter(obj)
  obj.canshoot = false
end
function states.backfall:update(obj, dt)
  local xv, yv = obj:getVelocity()
  local sx, sy = obj:getScale()
  obj:setVelocity(-50*sx, yv)
end
function states.backfall:hitground(obj, yv)
  obj:setState("landing")
end

states.rolling = {}
function states.rolling:update(obj, dt)
  obj.canshoot = false
  local sx, sy = obj:getScale()
  local xv, yv = obj:getVelocity()
  xv = sx*100
  obj:setVelocity(xv, yv)
  if obj.elapsed > obj.animation.total then
    obj:setVelocity(0, yv)
    obj:setState("landing")
  end
end
function states.rolling:fall(obj)
  obj:setState("falling")
end

states.dropdown = {}
function states.dropdown:enter(obj)
  obj.body:setFilter(cat.player, cat.pdefault - cat.platform, 0)
end
function states.dropdown:exit(obj)
  obj.body:setFilter(cat.player, cat.pdefault, 0)
end
function states.dropdown:update(obj, dt)
  if obj.elapsed >= obj.animation.total*2 then
    obj.body:setFilter(cat.player, cat.pdefault, 0)
    local e = obj.elapsed
    obj:setState("falling")
    obj.elapsed = e
  end
end
function states.dropdown:hitground(obj, yv)
  if yv > 110 then
    obj:setState("landing", yv)
  else
    obj:setState("standing")
  end
end
function states.dropdown:shoot(obj)
  local sx, sy = obj:getScale()
  obj:shoot(sx*2, -1, sx, 0)
end

states.falling = {}
function states.falling:update(obj, dt)
  obj.canshoot = true
  local xv, yv = obj:getVelocity()
  local xm, ym = obj:getMovement()
  --local sx, sy = obj:getScale()
  obj:setVelocity(50*xm, yv)
  if xm == -1 or xm == 1 then
    obj:setScale(xm, 1)
  end
end
function states.falling:hitground(obj)
  obj:setState("standing")
end
function states.falling:hitground(obj, yv)
  if yv > 110 then
    obj:setState("landing", yv)
  else
    if obj.elapsed <= obj.jumpstroke then
      obj.lastjump = -1
    end
    obj:setState("standing")
  end
end
function states.falling:climb(obj)
  obj:setState("climbing")
end
function states.falling:shoot(obj)
  local sx, sy = obj:getScale()
  obj:shoot(sx*2, -1, sx, 0)
end
function states.falling:leap(obj)
  if obj.prevstate == states.running then
    if obj.lvl.data.elapsed - obj.lastgrounded <= obj.jumpdelay then
      obj:setState("jumping")
    end
  end
end

states.jumping = {}
function states.jumping:enter(obj)
  obj:playSound("jump")
  obj.canshoot = true
  local xv, yv = obj:getVelocity()
  obj:setVelocity(xv, -200)
end
function states.jumping:exit(obj)
  obj.body:setGravityScale(1)
end
function states.jumping:update(obj, dt)
  local xv, yv = obj:getVelocity()
  local xm, ym, jump = obj:getMovement()
  xv = 50*xm
  obj:setVelocity(xv, yv)
  if xm == -1 or xm == 1 then
    obj:setScale(xm, 1)
  end
  if not jump then
    --obj.body:setGravityScale(3)
    obj.body:setGravityScale(5)
  end
  if yv > 0 then
    obj:setState("falling")
  end
end
function states.jumping:climb(obj)
  obj:setState("climbing")
end
--[[
function states.jumping:hitroof(obj)
  obj:setState("falling")
end
]]
function states.jumping:hitground(obj, yv)
  if yv > 110 then
    obj:setState("landing", yv)
  else
    obj:setState("standing")
  end
end
function states.jumping:shoot(obj)
  local sx, sy = obj:getScale()
  obj:shoot(sx*2, -1, sx, 0)
end

states.landing = {}
function states.landing:enter(obj)
  obj.canshoot = false
  obj:setVelocity(0, 0)
  obj:playSound("land")
  --local sx, sy = obj:getScale()
  --local anim = obj:spawn("animation", -2*sx, 2, 1, "smoke1")
  --anim:setColor(128, 128, 128)
end
function states.landing:update(obj, dt)
  if obj.elapsed >= obj.landingdelay then
    obj:setState("standing")
  end
end
function states.landing:fall(obj)
  obj:setState("falling")
end

states.exploding = {}
function states.exploding:enter(obj)
  obj.canshoot = false
  obj:playSound("die")
  obj:debree(6, 0, 0, 255, 255, 255)
end
function states.exploding:update(obj, dt)
  obj.body:setActive(false)
  if obj.elapsed >= 1 then
    local n = obj.name
    local lvl = obj.lvl
    local s = lvl.session
    lvl.destroyObject(obj)
    if s.lives > 0 then
      s.lives = s.lives - 1
      obj.checkpoint:join(n)
    end
    local d = lvl.data
    if d.controlled <= 1 then
      d.over = 3
      d.message = app.lang.gameover
      app.say("gameover")
      lvl.call("gameover")
    else
      local t = { "goodtry", "getready", "tryagain", "betterluck" }
      if s.lives == 0 then
        t = { "nomorelives", "finaltry", "lastchance" }
      end
      local n = table.random(t)
      app.say(n, 5)
    end
  end
end

return states