local animations =
{
  file = "player",
  running = { start = 1, frames = 5, cycle = "loop" },
  landing = { start = 6, frames = 3, cycle = "none" },
  standing = { start = 9, frames = 1, cycle = "loop" },
  waving = { start = 10, frames = 3, cycle = "bounce", steps = 6 },
  jumping = { start = 17, frames = 3, cycle = "none" },
  falling = { start = 20, frames = 3, cycle = "none" },
  dropdown = { start = 20, frames = 3, cycle = "none" },
  rolling = { start = 25, frames = 8, cycle = "loop" },
  backflip = { start = 33, frames = 8, cycle = "none" },
  backfall = { start = 41, frames = 8, cycle = "none" },
  hanging = { start = 49, frames = 8, cycle = "loop" },
  climbing = { start = 49, frames = 8, cycle = "loop" },
  exploding = { start = 66, frames = 7, cycle = "none" },
}

local parent = require("obj.agent")
local class = game.oo.class(parent)

class.states = require("obj.player-states")
class.animations = animations
class.landingdelay = 1/60*25
class.jumpstroke = 1/60*12
class.jumpdelay = 1/60*5
class.nosleep = true

function class:init(lvl, x, y, name)
  parent.init(self, lvl, x, y, 6)
  -- collision
  self.body:setLinearDamping(3)
  self.body:setBullet(true)
  self.shape = self.body:newBox(1.8, 2.6, 0, 1)
  self.shape:setFilterData(game.cat.player, game.cat.pdefault, 0)
  self.shape:setFriction(0)
  -- state
  self.elapsed = 0
  self.name = name
  self.cooldown = 0
  self.canshoot = true
  self.canclimb = true
  self.lastjump = -1
  self.lockout = -1
  self:setState("falling")
  local d = lvl.data
  d[name] = self
  d.controlled = d.controlled + 1
end

function class:destroy()
  local lvl = self.lvl
  local d = lvl.data
  d.controlled = d.controlled - 1
  local name = self.name
  if d[name] == self then
    d[name] = nil
  end
  self:setState()
  self.shape = nil
  self.elapsed = nil
  self.name = nil
  self.cooldown = nil
  self.canshoot = nil
  self.canclimb = nil
  self.lastjump = nil
  self.lockout = nil
  parent.destroy(self)
end

function class:getMovement()
  local mx, my = 0, 0
  local input = self.lvl.data[self.name.."a"]
  if input.left then
    mx = -1
  elseif input.right then
    mx = 1
  end
  if input.down then
    my = -1
  elseif input.up then
    my = 1
  end
  return mx, my, input.jump, input.shoot
end

function class:isNearLadder()
  local _, ym = self:getMovement()
  return self:getNearestLadder(ym < 0 and 4 or 8) ~= nil
end

function class:getNearestLadder(dist)
  local list = {}
  local contacts = self.body:getContacts()
  for i, contact in ipairs(contacts) do
    if contact:isTouching() and contact:isEnabled() then
      local f1, f2 = contact:getFixtures()
      local c1, c2 = f1:getFilterData(), f2:getFilterData()
      if c1 == game.cat.ladder then
        table.insert(list, f1)
      elseif c2 == game.cat.ladder then
        table.insert(list, f2)
      end
    end
  end
  local x1, y1 = self:getPosition()
  local f1, d1
  for i, f2 in ipairs(list) do
    local body = f2:getBody()
    local x2, y2 = body:getPosition()
    local d2 = y2 - y1
    if math.abs(d2) <= dist then
      if not d1 or d2 < d1 then
        f1, d1 = f2, d2
      end
    end
  end
  return f1, d1
end

function class:update(dt)
  parent.update(self, dt)
  self.body:setAwake(true)
  self.cooldown = self.cooldown - dt
  -- current movement, velocity and displacement
  local xm, ym, jump, shoot = self:getMovement()
  if self.canclimb and self:getNearestLadder(ym > 0 and 4 or 8) then
    if xm == 0 and ym ~= 0 and not jump then
      self:message("climb")
    end
  end
  local wasjumping = self.jumping
  local isjumping = jump
  self.jumping = jump
  local d = self.lvl.data
  if (isjumping and not wasjumping) then
    self.lastjump = d.elapsed
    self:message("leap")
  elseif d.elapsed - self.lastjump < self.jumpstroke then
    self:message("leap")
  end
end

function class:sync()
  local _, _, _, shoot = self:getMovement()
  local o = (self.canshoot and shoot) and 56 or 0
  --local o = (self.cooldown and self.cooldown > 0 and self.canshoot) and 56 or 0
  animations.running.start = 1 + o
  animations.standing.start = 9 + o
  animations.waving.start = 10 + o
  animations.jumping.start = 17 + o
  parent.sync(self)
end

function class:shoot(px, py, nx, ny)
  if self.cooldown > 0 then
    return
  end
  self.cooldown = 0.2
  self:spawn("bullet", px, py, nx, ny, game.cat.enemy)
end

function class:onDie(pts)
  self:setState("exploding")
  self:playSound("crush")
  game.screenshake(self.lvl, 0.25)
end

function class:collision(other, contact)
  local f1, f2 = contact:getFixtures()
  local b1, b2 = f1:getBody(), f2:getBody()
  local nx, ny = contact:getNormal()
  if b2 == self.body then
    b1, b2 = b2, b1
    nx, ny = -nx, -ny
  end
  local ax, ay = math.abs(nx), math.abs(ny)
  if self.state == self.states.rolling then
    if ax > ay then
      other:call("hit", 3)
    end
  else
    if ny > 0 then
      local ret = other:call("stomp", self, contact)
      if ret == false then
        self:message("jump")
      end
      return ret
    end
  end
end

function class:hit(pts)
  self:damage(pts)
end

function class:shot(pts)
  self:damage(pts)
end

return class