local parent = require("states.screen")
local class = parent:class()

local line = 10

function class:create(...)
  parent.create(self)
  self.elapsed = 0
  self.args = {...}
  self:layout(...)
end

function class:destroy()
  self.elapsed = nil
  self.args = nil
  parent.destroy(self)
end

function class:layout(n, ...)
  local prev = self.def and self.def.selected

  local fn = "layouts."..n
  package.loaded[fn] = nil
  local def = require(fn)
  def.selected = prev or def.selected
  self.def = def
  self:event("init", ...)
  def.selected = def.selected or 1

  for i, v in ipairs(def) do
    if type(v.value) == "number" then
      local j = math.floor(v.value + 0.5)
      assert(j >= 0 and j <= 10)
      local p1 = string.rep("║", j)
      local p2 = string.rep("═", 10 - j)
      v.text = "╠"..p1..p2.."╣ "..v.text
    elseif type(v.value) == "boolean" then
      local s = v.value and "☑" or "☐"
      v.text = app.lang.format(app.lang.map, v.text, s)
    end
  end
  
  local layer = self.layer
  layer:destroyChildren()
  local font = app.fonts.mini
  local w, h = self.view:getDimensions()
  local bw, bh = 0, line*#def/2
  
  self.sprites = {}
  for i, v in ipairs(def) do
    local txt = love.graphics.newText(font, v.text)
    bw = math.max(bw, txt:getWidth())
    local sprite = layer:newSprite(0, 0)
    sprite:setGraphic(txt)
    self.sprites[i] = sprite
  end
  for i, sprite in ipairs(self.sprites) do
    sprite:setPosition(-bw/2, -bh + (i - 1)*line)
  end
  bw = math.min(bw, w)
  def.w, def.h = bw, bh

  local verts = { {-w,-bh - 2},{w,-bh - 2},{w,bh + 2},{-w,bh + 2} }
  local mesh = love.graphics.newMesh(verts, "fan", "static")
  self.bg = layer:newSprite(0, 0)
  self.bg:setGraphic(mesh)
  self.bg:setIndex(1)

  local verts = { {-w,-bh},{w,-bh},{w,bh},{-w,bh} }
  local mesh = love.graphics.newMesh(verts, "fan", "static")
  self.fg = layer:newSprite(0, 0)
  self.fg:setGraphic(mesh)
  self.fg:setColorRGB(0, 0, 0)
  self.fg:setScale(0, 0)
  self.fg:setIndex(2)

  local index = self.def.selected
  local verts2 = { {-w,0},{w,0},{w,line},{-w,line} }
  local mesh2 = love.graphics.newMesh(verts2, "fan", "static")
  self.cursor = layer:newSprite(0, 0)
  self.cursor:setPosition(0, index*line - bh)
  self.cursor:setGraphic(mesh2)
  self.cursor:setColorRGB(255, 0, 0)
  self.cursor:setIndex(3)

  self.elapsed = 0
end

function class:update(dt)
  self:event("update", dt)
  local w, h = self.view:getDimensions()
  local bw, bh = w, line*#self.def/2
  local index = self.def.selected
  self.cursor:setPosition(0, (index - 1)*line - bh)
  self.elapsed = self.elapsed + dt
  local r = math.min(self.elapsed/0.2, 1)
  self.fg:setScale(1, r)
  for i, v in ipairs(self.sprites) do
    v:setAlpha(self.def[i].disabled and 0.5 or 1)
  end
  self.cursor:setVisible(self.elapsed >= 0.2)
  for i, s in ipairs(self.sprites) do
    s:setVisible(r == 1)
  end
end

function class:reset()
  self:layout(unpack(self.args))
end

function class:attach(id)
  self:reset()
end

function class:detach(id)
  self:reset()
end

function class:event(e, ...)
  local func = self.def[e]
  if func then
    func(...)
  end
end

function class:select(dt)
  local i = self.def.selected + dt
  if i >= 1 and i <= #self.def then
    if i ~= self.def.selected then
      app.playSound("select")
      self.def.selected = i
      self:event("select", i)
    end
  end
end

function class:up()
  self:select(-1)
end

function class:down()
  self:select(1)
end

function class:left()
  self:toggle(-1)
end

function class:right()
  self:toggle(1)
end

function class:start()
  local d = self.def
  local i = d.selected
  if not d[i].disabled then
    self:event("choose", i)
    app.playSound("start")
  else
    app.playSound("error")
  end
end

function class:toggle(dt)
  local d = self.def
  local i = d.selected
  local v = self.def[i]
  if v and type(v.value) == "number" then
    local j = v.value + dt
    j = math.floor(j + 0.5)
    if j >= 0 and j <= 10 then
      app.playSound("select")
      self:event("choose", i, j)
    end
  end
end

function class:shoot()
  self:start()
end

function class:jump()
  self:start()
end

function class:cancel()
  app.playSound("error")
  self:event("cancel")
end

return class