local reg = debug.getregistry()
local lp = love.physics

local _bodies = {}
local _joints = {}
local _bodiesR = setmetatable({}, { __mode = "k" })
local _jointsR = setmetatable({}, { __mode = "k" })

-- World
function reg.World:index()
  if _bodies[self] == nil then
    _bodies[self] = {}
    _joints[self] = {}
  end
end

function reg.World:deindex()
  if _bodies[self] ~= nil then
    _bodies[self] = nil
    _joints[self] = nil
  end
end

function reg.World:getBody(id)
  return _bodies[self][id]
end

function reg.World:getBodyID(b)
  return _bodiesR[b]
end

function reg.World:setBodyID(b, id)
  local pid = _bodiesR[b]
  if pid then
    _bodies[self][pid] = nil
  end
  _bodiesR[b] = id
  if id then
    assert(not _bodies[self][id], "id already taken")
    _bodies[self][id] = b
  end
end

function reg.World:nextID()
  local b = _bodies[self]
  for i = 1, math.huge do
    if b[i] == nil then
      return i
    end
  end
end

function reg.World:getJoint(id)
  return _joints[self][id]
end

function reg.World:getJointID(j)
  return _jointsR[j]
end

function reg.World:setJointID(id, j)
  if j then
    assert(_joints[self][id] == nil, "joint id already taken")
    assert(_jointsR[j] == nil, "joint exists with another id")
    _jointsR[j] = id
  end
  _joints[self][id] = j
end

local newBody = lp.newBody
function reg.World:newBody(x, y, t)
  return newBody(self, x, y, t)
end

local destroyBody = reg.Body.destroy
function reg.World:destroyBody(b)
  destroyBody(b)
end

local destroyJoint = reg.Joint.destroy
function reg.World:destroyJoint(j)
  destroyJoint(j)
end

-- Body
local newPolygonShape = lp.newPolygonShape
local newFixture = lp.newFixture
function reg.Body:newPolygon(...)
  local s = newPolygonShape(...)
  return newFixture(self, s)
end

local newRectangleShape = lp.newRectangleShape
function reg.Body:newBox(w, h, lx, ly, la)
  lx, ly = lx or 0, ly or 0
  la = la or 0
  local s = newRectangleShape(lx, ly, w*2, h*2, la)
  return newFixture(self, s)
end

local newCircleShape = lp.newCircleShape
function reg.Body:newCircle(radius, lx, ly)
  lx, ly = lx or 0, ly or 0
  local s = newCircleShape(lx, ly, radius)
  return newFixture(self, s)
end

local newChainShape = lp.newChainShape
function reg.Body:newChain(vertices, loop)
  local s = newChainShape(loop, vertices)
  return newFixture(self, s)
end

local newEdgeShape = lp.newEdgeShape
function reg.Body:newEdge(x1, y1, x2, y2)
  local s = newEdgeShape(x1, y1, x2, y2)
  return newFixture(self, s)
end

local destroyFixture = reg.Fixture.destroy
function reg.Body:destroyFixture(f)
  destroyFixture(f)
end

local getFixtures = reg.Body.getFixtures
function reg.Body:destroyFixtures()
  local fixtures = getFixtures(self)
  for _, f in ipairs(fixtures) do
    destroyFixture(f)
  end
end

local getJoints = reg.Body.getJoints
local getJointType = reg.Joint.getType
function reg.Body:findJoint(t)
  local joints = getJoints(self)
  for _, j in ipairs(joints) do
    if getJointType(j) == t then
      return j
    end
  end
end

local setFilterData = reg.Fixture.setFilterData
function reg.Body:setFilter(c, m, g)
  local fixtures = getFixtures(self)
  for _, f in ipairs(fixtures) do
    setFilterData(f, c, m, g)
  end
end

local getContacts = reg.Body.getContacts
local setFriction = reg.Fixture.setFriction
local setContacFriction = reg.Contact.setFriction
function reg.Body:setFriction(v)
  local fixtures = getFixtures(self)
  for _, f in ipairs(fixtures) do
    setFriction(f, v)
  end
  local contacts = getContacts(self)
  for _, c in ipairs(contacts) do
    setContacFriction(c, v)
  end
end

local getBoundingBox = reg.Fixture.getBoundingBox
function reg.Body:getBoundingBox()
  -- get the aabb of the body
  local fixtures = getFixtures(self)
  if #fixtures > 0 then
    local x1, y1, x2, y2 = getBoundingBox(fixtures[1])
    for i = 2, #fixtures do
      local x3, y3, x4, y4 = getBoundingBox(fixtures[i])
      x1 = (x3 < x1) and x3 or x1
      x2 = (x4 > x2) and x4 or x2
      y1 = (y3 < y1) and y3 or y1
      y2 = (y4 > y2) and y4 or y2
    end
    return x1, y1, x2, y2
  end
end

local testPoint = reg.Fixture.testPoint
function reg.Body:testPoint(x, y)
  -- iterate fixtures
  local fixtures = getFixtures(self)
  for _, f in ipairs(fixtures) do
    if testPoint(f, x, y) then
      return true
    end
  end
  return false
end

local getContactFixtures = reg.Contact.getFixtures
local getBody = reg.Fixture.getBody
function reg.Body:getOther(contact)
  local a, b = getContactFixtures(contact)
  local other = getBody(a)
  if other == self then
    other = getBody(b)
  end
  return other
end

local isContactTouching = reg.Contact.isTouching
local getContacts = reg.Body.getContacts
function reg.Body:isTouchingFixture(other)
  -- iterate possible collisions
  local contacts = getContacts(self)
  for _, c in ipairs(contacts) do
    if isContactTouching(c) then
      local fa, fb = getContactFixtures(c)
      if other == fa or other == fb then
        return true
      end
    end
  end
  return false
end

function reg.Body:isTouchingOther(other)
  if other:type() == "Body" then
    return self:isTouching(other)
  end
  return self:isTouchingFixture(other)
end

-- Fixture
reg.Fixture.setFilter = reg.Fixture.setFilterData

-- Shape
reg.CircleShape.getLocalPosition = reg.CircleShape.getPoint

reg.PolygonShape.getVertices = reg.PolygonShape.getPoints
reg.ChainShape.getVertices = reg.ChainShape.getPoints
reg.EdgeShape.getVertices = reg.EdgeShape.getPoints

-- Joint
local newRevoluteJoint = lp.newRevoluteJoint
function reg.World:newRevoluteJoint(a, b, x, y, cc)
  return newRevoluteJoint(a, b, x, y, cc)
end

reg.RevoluteJoint.getJointPosition = reg.RevoluteJoint.getJointAngle
reg.RevoluteJoint.enableMotor = reg.RevoluteJoint.setMotorEnabled
reg.RevoluteJoint.isLimitEnabled = reg.RevoluteJoint.areLimitsEnabled

local newPrismaticJoint = lp.newPrismaticJoint
function reg.World:newPrismaticJoint(a, b, x, y, ax, ay, cc)
  return newPrismaticJoint(a, b, x, y, ax, ay, cc)
end

local getBodies = reg.Joint.getBodies
local getAnchors = reg.Joint.getAnchors
local getJointTranslation = reg.PrismaticJoint.getJointTranslation
local getAxis = reg.PrismaticJoint.getAxis
local getPosition = reg.Body.getPosition
local setPosition = reg.Body.setPosition
function reg.PrismaticJoint:setJointTranslation(t2)
  -- note affects the second body only!
  local _, b = getBodies(self)
  local x, y = getPosition(b)
  local _, _, x2, y2 = getAnchors(self)
  -- offset anchor by the body's position
  x2, y2 = x2 - (x2 - x), y2 - (y2 - y)
  -- find the origin, where translation = 0
  local t1 = getJointTranslation(self)
  local nx, ny = getAxis(self)
  local ox, oy = x2 - nx*t1, y2 - ny*t1
  -- move the body to desired translation
  setPosition(b, ox + nx*t2, oy + ny*t2)
end

reg.PrismaticJoint.getJointPosition = reg.PrismaticJoint.getJointTranslation
reg.PrismaticJoint.enableMotor = reg.PrismaticJoint.setMotorEnabled
reg.PrismaticJoint.isLimitEnabled = reg.PrismaticJoint.areLimitsEnabled

local newDistanceJoint = lp.newDistanceJoint
function reg.World:newDistanceJoint(a, b, p1x, p1y, p2x, p2y, cc)
  return newDistanceJoint(a, b, p1x, p1y, p2x, p2y, cc)
end

local newRopeJoint = lp.newRopeJoint
local sqrt = math.sqrt
function reg.World:newRopeJoint(a, b, p1x, p1y, p2x, p2y, l, cc)
  if l == nil then
    local lx, ly = p1x - p2x, p1y - p2y
    l = sqrt(lx*lx + ly*ly)
  end
  return newRopeJoint(a, b, p1x, p1y, p2x, p2y, l, cc)
end

local newPulleyJoint = lp.newPulleyJoint
function reg.World:newPulleyJoint(a, b, p1x, p1y, p2x, p2y, p3x, p3y, p4x, p4y, ratio, cc)
  return newPulleyJoint(a, b, p1y, p2x, p2y, p3x, p3y, p4x, p4y, ratio, cc)
end

local newGearJoint = lp.newGearJoint
function reg.World:newGearJoint(a, b, j1, j2, ratio, cc)
  return newGearJoint(j1, j2, ratio, cc)
end

local newWeldJoint = lp.newWeldJoint
function reg.World:newWeldJoint(a, b, p1x, p1y, cc)
  return newWeldJoint(a, b, p1x, p1y, cc)
end

local newFrictionJoint = lp.newFrictionJoint
function reg.World:newFrictionJoint(a, b, p1x, p1y, cc)
  return newFrictionJoint(a, b, p1x, p1y, cc)
end

local newWheelJoint = lp.newWheelJoint
function reg.World:newWheelJoint(a, b, p1x, p1y, ax, ay, cc)
  return newWheelJoint(a, b, p1x, p1y, ax, ay, cc)
end

local newMouseJoint = lp.newMouseJoint
function reg.World:newMouseJoint(a, x, y, mf)
  local joint = newMouseJoint(a, x, y)
  joint:setMaxForce(mf)
  return joint
end

lp.setMeter(1)