local function get(x, t, s)
  for i = 1, #x do
    local v = x[i]
    if v[0] == t then
      if s then
        return v[s]
      end
      return v
    end
  end
end

local function getPath(points)
  local path = {}
  for i = 1, #points do
    local v = points[i]
    local j = (i - 1)*2
    path[j + 1] = v.x
    path[j + 2] = v.y
  end
  return path
end

local b2d = debug.getregistry()
local lp = love.physics

--- World
function b2d.World:load(w)
  self:index()
  for i = 1, #w do
    local v = w[i]
    local tag = v[0]
    if tag == "G" then
      self:setGravity(v.x, v.y)
    elseif tag == "B" then
      local b = self:loadBody(v)
      self:setBodyID(b, v.id)
    elseif tag == "J" then
      local j = self:loadJoint(v)
      self:setJointID(v.id, j)
    end
  end
end

function b2d.World:unload()
  self:deindex()
  self:destroy()
end

--- Body

local bodyTypes =
{
  s = "static",
  d = "dynamic",
  k = "kinematic"
}

function b2d.World:loadBody(b)
  local t = bodyTypes[b.type]
  local body = self:newBody(b.x, b.y, t)
  if b.a then
    body:setAngle(b.a)
  end
  for i = 1, #b do
    local v = b[i]
    local t = v[0]
    if t == "F" then
      body:loadFixture(v)
    elseif t == "LV" then
      body:setLinearVelocity(v.x, v.y)
    elseif t == "AV" then
      body:setAngularVelocity(v[1])
    elseif t == "LD" then
      body:setLinearDamping(v[1])
    elseif t == "AD" then
      body:setAngularDamping(v[1])
    elseif t == "GS" then
      body:setGravityScale(v[1])
    elseif t == "FR" then
      body:setFixedRotation(v[1] == true)
    elseif t == "IB" then
      body:setBullet(v[1] == true)
    elseif t == "SA" then
      body:setSleepingAllowed(v[1] == true)
    elseif t == "AW" then
      body:setAwake(v[1] == true)
    elseif t == "IA" then
      body:setActive(v[1] == true)
    elseif t == "SS" then
      -- scale
    else
      error("unknown tag in body:" .. (b.id or "nil"))
    end
  end
  body:resetMassData()
  return body
end

--- Fixtures
function b2d.Body:loadFixture(f)
  local t = f.type
  local fixture
  if t == "c" then
    -- circle
    local radius = get(f, "R", 1)
    local lp = get(f, "LP")
    local lpx, lpy = 0, 0
    if lp then
      lpx, lpy = lp.x, lp.y
    end
    fixture = self:newCircle(radius, lpx, lpy)
  elseif t == "p" then
    -- convex polygon
    local path = getPath(f)
    fixture = self:newPolygon(path)
  elseif t == "h" then
    -- chain
    local path = getPath(f)
    fixture = self:newChain(path, f.l == true)
  elseif t == "e" then
    -- edge
    local v1 = get(f, "V1")
    local v2 = get(f, "V2")
    fixture = self:newEdge(v1.x, v1.y, v2.x, v2.y)
  else
    error("unknown fixture type: " .. t)
  end
  -- fixture
  fixture:setDensity(f.d or 1)
  fixture:setFriction(f.f or 0)
  fixture:setRestitution(f.r or 0)
  fixture:setSensor(f.s == true)
  -- filter
  fixture:setFilter(f.c or 0x0001, f.m or 0xffff, f.g or 0)

  return fixture
end

--- Joints
function b2d.World:loadJoint(j)
  local t = j.type
  local jid = j.id or 'nil'
  -- generic joint properties
  local cc = (j.cc == true)
  local bodyA = self:getBody(j.body1)
  local bodyB = self:getBody(j.body2)
  local joint
  if t == "g" then
    local j1 = get(j, "J1", 1)
    local j2 = get(j, "J2", 1)
    local joint1 = self:getJoint(j1)
    local joint2 = self:getJoint(j2)
    if joint1 == nil then
      error("joint1 is nil in joint:" .. jid)
    end
    if joint2 == nil then
      error("joint2 is nil in joint:" .. jid)
    end
    --[[
    if joint1 == joint2 then
      error("joint1 == joint2 in joint:" .. jid)
    end
    ]]
    local ro = get(j, "RO", 1) or 1 -- ratio
    joint = lp.newGearJoint(joint1, joint2, ro, cc)
  else
    -- local anchors
    local a1 = get(j, "A1")
    local a2 = get(j, "A2")
    local a1x, a1y = bodyA:getWorldPoint(a1.x, a1.y)
    local a2x, a2y = bodyB:getWorldPoint(a2.x, a2.y)
    if t == "o" then
      local ml = get(j, "ML", 1) or 0 -- max length
      --if maxLength < b2d.linearSlop then
        --error("max length < linearSlop in joint:" .. jid)
      --end
      joint = lp.newRopeJoint(bodyA, bodyB, a1x, a1y, a2x, a2y, ml, cc)
    elseif t == "w" then
      local ra = get(j, "RA", 1) or 0 -- reference angle
      joint = lp.newWeldJoint(bodyA, bodyB, a1x, a1y, a2x, a2y, cc, ra)
      local fq = get(j, "HZ", 1) or 0 -- frequency in Hertz
      joint:setFrequency(fq)
      local dr = get(j, "DR", 1) or 0 -- damping ratio
      joint:setDampingRatio(dr)
    elseif t == "u" then
      -- ground anchors
      local g1 = get(j, "G1")
      local g2 = get(j, "G2")
      local ra = get(j, "RO", 1) or 1 -- ratio
      joint = lp.newPulleyJoint(bodyA, bodyB, g1.x, g1.y, g2.x, g2.y, a1x, a1y, a2x, a2y, ra, cc)
      --local lA = get(j, "L1", 1) or 0 -- length A
      --local lB = get(j, "L2", 1) or 0 -- length B
      --[[
      local ml1 = get(j, "ML1", 1) or 0 -- max length 1
      local ml2 = get(j, "ML2", 1) or 0 -- max length 2
      joint:setMaxLengths(ml1, ml2)
      ]]
    elseif t == "r" then
      local ra = get(j, "RA", 1) or 0 -- reference angle
      joint = lp.newRevoluteJoint(bodyA, bodyB, a1x, a1y, a2x, a2y, cc, ra)
      local m = get(j, "MT")
      if m then
        joint:setMaxMotorTorque(m.t or 0)
        joint:setMotorSpeed(m.s or 0)
        joint:setMotorEnabled(m.e == true)
      end
      local l = get(j, "LT")
      if l then
        local lo = l.l or 0
        local hi = l.u or 0
        joint:setLimits(lo, hi)
        joint:setLimitsEnabled(l.e == true)
      end
    elseif t == "p" then
      -- local axis
      local la = get(j, "LA") -- local axis
      local lax, lay = bodyA:getWorldVector(la.x, la.y)
      local ra = get(j, "RA", 1) or 0 -- reference angle
      joint = lp.newPrismaticJoint(bodyA, bodyB, a1x, a1y, a2x, a2y, lax, lay, cc, ra)
      local l = get(j, "LT")
      if l then
        local lo = l.l or 0
        local hi = l.u or 0
        joint:setLimits(lo, hi)
        joint:setLimitsEnabled(l.e == true)
      end
      local m = get(j, "MT")
      if m then
        joint:setMaxMotorForce(m.f or 0)
        joint:setMotorSpeed(m.s or 0)
        joint:setMotorEnabled(m.e == true)
      end
    elseif t == "d" then
      joint = lp.newDistanceJoint(bodyA, bodyB, a1x, a1y, a2x, a2y, cc)
      local ln = get(j, "L", 1) or 0 -- length
      local fq = get(j, "HZ", 1) or 0 -- frequency in Hertz
      local dr = get(j, "DR", 1) or 0 -- damping ratio
      joint:setLength(ln)
      joint:setFrequency(fq)
      joint:setDampingRatio(dr)
      --if length < b2d.linearSlop then
        --error("length < linearSlop in joint:" .. jid)
      --end
    else
      error("unsupported type in joint:" .. jid)
    end
  end

  return joint
end