--- Custom formatting functions
local utf8 = require("utf8")
local path = (...):match("(.-)[^%.]+$")
local toupper = require(path.."upper")

local text = {}

-- Uppercase
function text.upper(s)
  return string.gsub(s, utf8.charpattern, toupper)
end

-- String substitution
function text.sub(s, i, j)
  local n = utf8.len(s)
  if i > n then
    i = n
  end
  if not j or j > n then
    j = n
  end
  local j1 = utf8.offset(s, i)
  local j2 = utf8.offset(s, j + 1) - 1
  return string.sub(s, j1, j2)
end

-- String length
function text.len(s)
  return utf8.len(s)
end

-- String reverse
function text.reverse(s)
  local t = {}
  for p, c in utf8.codes(s) do
    table.insert(t, 1, c)
  end
  return utf8.char(unpack(t))
end

-- Formats a string, using numerically ordered parameters
-- This is useful for handling both LTR and RTL languages
function text.format(s, ...)
  local t = {...}
  for i, v in ipairs(t) do
    local k = tostring(i)
    t[k] = v
  end
  return string.gsub(s, "$(%d+)", t)
end

-- Formats time with second precision, for example 95 => 01:35
function text.seconds(t)
  local s = math.floor(t)%60
  local m = math.floor(t/60)%60
  local out = string.format("%02d:%02d", m, s)
  if t >= 3600 then
    local h = math.floor(t/3600)
    out = string.format("%d:%s", h, out)
  end
  return out
end

-- Format time with millisecond precision, for example 95.123 => 01:35.123
function text.milliseconds(t)
  local out = text.seconds(t/1000)
  local ms = math.floor(t)%1000
  return string.format("%s.%03d", out, ms)
end

-- Groups digits by three
function text.number(n, q)
  -- separator
  q = q or ","
  -- no scientific notation
  n = string.format("%.f", n)
  -- black voodoo
	local l, n, r = string.match(n, '^([^%d]*%d)(%d*)(.-)$')
  local s = n:reverse():gsub('(%d%d%d)','%1'..q):reverse()
	return l..s..r
end

-- Replaces the numerals in a string to Hindu-Arabic
local nums =
{
  ["0"]="٠",["1"]="١",["2"]="٢",["3"]="٣",["4"]="٤",
  ["5"]="٥",["6"]="٦",["7"]="٧",["8"]="٨",["9"]="٩"
}
function text.numberha(n)
  n = string.format("%.f", n)
  return string.gsub(n, "%d", nums)
end

-- Hebrew
function text.ishebrew(s)
  for p, c in utf8.codes(s) do
    if c >= 0x0590 and c <= 0x05FF then
      return true
    end
  end
end

-- Arabic
function text.isarabic(s)
  for p, c in utf8.codes(s) do
    if c >= 0x0600 and c <= 0x06FF then
      return true
    end
    if c >= 0x0750 and c <= 0x077F then
      return true
    end
    if c >= 0x08A0 and c <= 0x08FF then
      return true
    end
    if c >= 0xFB50 and c <= 0xFDFF then
      return true
    end
    if c >= 0xFE70 and c <= 0xFEFF then
      return true
    end
    if c >= 0x10E60 and c <= 0x10E7F then
      return true
    end
    if c >= 0x1EC70 and c <= 0x1ECBF then
      return true
    end
    if c >= 0x1EE00 and c <= 0x1EEFF then
      return true
    end
  end
end

--- Reverse RTL text so that it renders properly with Love2D
function text.rtl(s)
  local t = {}
  for word, space in s:gmatch("(%S+)([%W%P]*)") do
    if text.ishebrew(word) or text.isarabic(word) then
      word = text.reverse(word)
    end
    table.insert(t, 1, word)
    table.insert(t, 1, space)
  end
  return table.concat(t)
end

return text