local tinsert = table.insert
local tremove = table.remove

local layer = {}
local layerMT = { __index = layer }

local reg = debug.getregistry()
reg.Layer = layer

setmetatable(layer, { __index = reg.Node })

local lg = love.graphics

function layer.new(x, y, mt)
  local t = reg.Node.new(x, y, mt or layerMT)
  t.list = {}
  return t
end

function layer:destroy()
  self:destroyChildren()
  self.list = nil
  reg.Node.destroy(self)
end

function layer:destroyChildren()
  local list = self.list
  if list then
    for i = #list, 1, -1 do
      local v = list[i]
      v.parent = nil
      v:destroy()
      list[i] = nil
    end
  end
end

function layer:newSprite(x, y)
  local c = reg.Sprite.new(x, y)
  self:insertChild(c)
  return c
end

function layer:newLayer(x, y)
  local c = reg.Layer.new(x, y)
  self:insertChild(c)
  return c
end

-- Hierarchy
function layer:insertChild(c)
  local p = c.parent
  if p then
    p:removeChild(c)
  end
  c.parent = self
  tinsert(self.list, c)
end

function layer:removeChild(c)
  local list = self.list
  for i = 1, #list do
    if c == list[i] then
      c.parent = nil
      tremove(list, i)
      break
    end
  end
end

function layer:getChild(i)
  return self.list[i]
end

function layer:getChildIndex(c)
  local list = self.list
  for i = 1, #list do
    if list[i] == c then
      return i
    end
  end
end

function layer:setChildIndex(c, i)
  local j = self:getChildIndex(c)
  if not j then
    return
  end
  local list = self.list
  if i == j or i < 1 or i > #list then
    return
  end
  tremove(list, j)
  tinsert(list, i, c)
end

function layer:sort(func)
  table.sort(self.list, func)
end

function layer:clear()
  local list = self.list
  for i = #list, 1, -1 do
    list[i].parent = nil
    list[i] = nil
  end
end

-- Rendering

function layer:draw()
  if not self.visible then
    return
  end
  local trans = self.transform
  if self.changed then
    trans:setTransformation(self.x, self.y, self.r, self.sx, self.sy)
    self.changed = nil
  end
  lg.push("transform")
  lg.applyTransform(trans)
  for i, v in ipairs(self.list) do
    v:draw()
  end
  lg.pop()
end

return layer.new