local node = {}
local nodeMT = { __index = node }

local reg = debug.getregistry()
reg.Node = node

function node.new(x, y, mt)
  --assert(x and y)
  local t = { x = x, y = y, r = 0, sx = 1, sy = 1 }
  t.transform = love.math.newTransform(x, y)
  t.visible = true
  t.changed = true
  return setmetatable(t, mt or nodeMT)
end

function node:destroy()
  local p = self.parent
  if p then
    p:removeChild(self)
  end
  self.transform = nil
end

-- Hierarchy
function node:getRoot()
  local p = self.parent
  if not p then
    return self
  end
  return p:getRoot()
end

function node:setParent(p2)
  local p1 = self.parent
  if p1 then
    p1:removeChild(self)
  end
  p2:insertChild(self)
end

-- Depth
function node:getIndex()
  local p = self.parent
  return p and p:getChildIndex(self)
end

function node:setIndex(i)
  local p = self.parent
  return p and p:setChildIndex(self, i)
end

-- Transformation
function node:getPosition()
  return self.x, self.y
end

function node:setPosition(x, y)
  self.x = x
  self.y = y
  self.changed = true
end

function node:getRotation()
  return self.r
end

function node:setRotation(r)
  self.r = r
  self.changed = true
end

function node:getScale()
  return self.sx, self.sy
end

function node:setScale(sx, sy)
  self.sx = sx
  self.sy = sy
  self.changed = true
end

function node:getTransform()
  return self.x, self.y, self.r
end

function node:setTransform(x, y, r)
  self.x = x
  self.y = y
  self.r = r
  self.changed = true
end

-- Visibility
function node:getVisible()
  return self.visible
end

function node:setVisible(on)
  self.visible = (on == true)
end

-- Conversion
function node:screenToLocal(x, y)
  local r = self:getRoot()
  if r and r.screenToScene then
    x, y = r:screenToScene(x, y)
    return self:sceneToLocal(x, y)
  end
end

function node:localToScreen(x, y)
  local r = self:getRoot()
  if r and r.sceneToScreen then
    x, y = self:localToScene(x, y)
    return r:sceneToScreen(x, y)
  end
end

function node:sceneToLocal(x, y)
  local p = self.parent
  if p then
    x, y = p:sceneToLocal(x, y)
  end
  return self:parentToLocal(x, y)
end

function node:localToScene(x, y)
  x, y = self:localToParent(x, y)
  local p = self.parent
  if p then
    x, y = p:localToScene(x, y)
  end
  return x, y
end

function node:parentToLocal(x, y)
  -- translate
  x = x - self.x
  y = y - self.y
  -- scale
  x = x*self.sx
  y = y*self.sy
  -- rotate
  local r = -self.r
  local c = math.cos(r)
  local s = math.sin(r)
  local rx = c*x - s*y
  local ry = s*x + c*y
  x, y = rx, ry
  return x, y
end

function node:localToParent(x, y)
  -- rotate
  local r = self.r
  local c = math.cos(r)
  local s = math.sin(r)
  local rx = c*x - s*y
  local ry = s*x + c*y
  x, y = rx, ry
  -- scale
  x = x/self.sx
  y = y/self.sy
  -- translate
  x = x + self.x
  y = y + self.y
  return x, y
end

return node.new