local view = {}
local viewMT = { __index = view }

local reg = debug.getregistry()
reg.View = view

setmetatable(view, { __index = reg.Layer })

local lg = love.graphics

function view.new(vx, vy, vw, vh)
  if vx == nil then
    vx, vy = 0, 0
    vw, vh = lg.getDimensions()
  end
  local t = reg.Layer.new(0, 0, viewMT)
  t.vx, t.vy = vx, vy
  t.vw, t.vh = vw, vh
  t.background = { 0, 0, 0, 1 }
  return t
end

function view:destroy()
  self.background = nil
  reg.Layer.destroy(self)
end

function view:setBounds(vx, vy, vw, vh)
  self.vx, self.vy = vx, vy
  vw = math.ceil(vw)
  vh = math.ceil(vh)
  if self.vw == vw and self.vh == vh then
    return
  end
  self.vw, self.vh = vw, vh
  if self.canvas then
    local ok, canvas = pcall(lg.newCanvas, vw, vh)
    self.canvas = ok and canvas
  end
end

function view:getBounds()
  return self.vx, self.vy, self.vw, self.vh
end

function view:getDimensions()
  return self.vw, self.vh
end

function view:setDimensions(vw, vh)
  self:setBounds(self.vx, self.vy, vw, vh)
end

function view:setScene(x, y, w, h)
  self.x = x
  self.y = y
  if w and h then
    self.sx = self.vw/w
    self.sy = self.vh/h
  end
end

function view:getScene()
  local w, h = self.vw/self.sx, self.vh/self.sy
  return self.x, self.y, w, h
end

function view:setBackground(r, g, b, a)
  if type(r) == "table" then
    r, g, b, a = unpack(r)
  end
  a = a or 1
  local bg = self.background
  bg[1], bg[2], bg[3], bg[4] = r, g, b, a
end

function view:getBackground()
  local bg = self.background
  return bg[1], bg[2], bg[3], bg[4]
end

function view:setBackgroundRGB(r, g, b, a)
  if type(r) == "table" then
    r, g, b, a = unpack(r)
  end
  a = a or 255
  local bg = self.background
  bg[1], bg[2], bg[3], bg[4] = r/255, g/255, b/255, a/255
end

function view:getBackgroundRGB()
  local bg = self.background
  return bg[1]*255, bg[2]*255, bg[3]*255, bg[4]*255
end

-- Rendering
function view:setShader(shader)
  if self.shader == shader then
    return
  end
  self.shader = shader
  if shader and not self.canvas then
    local ok, canvas = pcall(lg.newCanvas, self.vw, self.vh)
    self.canvas = ok and canvas
  end
end

function view:getShader()
  return self.shader
end

function view:draw()
  if not self.visible then
    return
  end

  local vx, vy, vw, vh = self:getBounds()
  local canvas = self.canvas
  local shader = self.shader

  lg.origin()
  lg.setBlendMode("alpha", "alphamultiply")
  if canvas and shader then
    lg.setCanvas(canvas)
    lg.clear(self.background)
  else
    lg.setScissor(vx, vy, vw, vh)
    lg.setColor(self.background)
    lg.rectangle("fill", vx, vy, vw, vh)
  end
  lg.push("transform")
  if not canvas then
    lg.translate(vx, vy)
  end
  lg.translate(vw/2, vh/2)
  lg.scale(self.sx, self.sy)
  lg.rotate(self.r)
  lg.translate(-self.x, self.y)
  for i, v in ipairs(self.list) do
    v:draw()
  end
  lg.pop()

  if canvas and shader then
    lg.setCanvas()
    lg.reset()
    lg.setShader(shader)
    lg.draw(canvas, vx, vy)
    lg.setShader()
  else
    lg.setScissor()
  end
end

-- Conversion
function view:localToScene(x, y)
  return x, y
end

function view:sceneToLocal(x, y)
  return x, y
end

function view:screenToScene(x, y)
  -- origin (center of the viewport)
  x = x - self.vw/2
  y = y - self.vh/2
  -- flip (y-axis increases up)
  y = -y
  -- transform
  x, y = self:localToParent(x, y)
  return x, y
end

function view:sceneToScreen(x, y)
  -- transform
  x, y = self:parentToLocal(x, y)
  -- flip (y-axis increases down)
  y = -y
  -- origin (top left of the window)
  x = self.vw/2 + x
  y = self.vh/2 + y
  return x, y
end

return view.new